﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace WPvP
{
    //this class computes a few stuff about combat handling
    public class Behavior
    {
        private static List<double> manaPercentsAfterCombats;
        private static List<double> healthPercentsAfterCombats;

        private static int WINDOW = 4;

        public static void Reset()
        {
            WPvPRoutine.Debug("Resetting oracle data...");
            WPvPRoutine.maxAdds = 1;
            manaPercentsAfterCombats = new List<double>();
            healthPercentsAfterCombats = new List<double>();
        }

        public static void AddCombatStats()
        {
            if (WPvPRoutine.Me.IsDead || WPvPRoutine.Me.IsGhost)
            {
                WPvPRoutine.Debug("I died. resetting oracle.");
                Reset();
                return;
            }

            if (manaPercentsAfterCombats.Count >= WINDOW)
                manaPercentsAfterCombats.RemoveAt(0);
            if (healthPercentsAfterCombats.Count >= WINDOW)
                healthPercentsAfterCombats.RemoveAt(0);

            healthPercentsAfterCombats.Add(WPvPRoutine.Me.HealthPercent);
            manaPercentsAfterCombats.Add(WPvPRoutine.Me.HealthPercent);

            SetMaxAdds();
        }

        public static void SetMaxAdds()
        {
            if (manaPercentsAfterCombats.Count == WINDOW)
            {
                double worstDeltaMana = 0;
                double worstDeltaHealth = 0;

                double averageMana = 0, averageHealth = 0;

                foreach (var p in manaPercentsAfterCombats)
                {
                    averageMana += p / WINDOW;
                    foreach (var p2 in manaPercentsAfterCombats)
                        if (Math.Abs(p - p2) > worstDeltaMana)
                            worstDeltaMana = Math.Abs(p - p2);
                }

                foreach (var p in healthPercentsAfterCombats)
                {
                    averageHealth += p / WINDOW;
                    foreach (var p2 in healthPercentsAfterCombats)
                        if (Math.Abs(p - p2) > worstDeltaHealth)
                            worstDeltaHealth = Math.Abs(p - p2);
                }

                WPvPRoutine.Debug("Oracle: deltaMana={0}, deltaLife={1}", worstDeltaMana, worstDeltaHealth);
                WPvPRoutine.Debug("Oracle: averageMana={0}, averageLife={1}", averageMana, averageHealth);

                //if stats are acceptable, consider pulling more mobs

                var worstDelta = Math.Max(worstDeltaHealth, worstDeltaMana);
                var worstAverage = Math.Max(averageMana, averageHealth);

                if (worstAverage > 75 && worstDelta < 10)
                {
                    WPvPRoutine.Debug("Nice stats! Trying to pull more mobs from now on.");
                    WPvPRoutine.maxAdds = Math.Min(3, WPvPRoutine.maxAdds + 1); ;
                }
                else
                {
                    WPvPRoutine.Debug("Dangerous combat behavior. Lowering max Adds!");
                    WPvPRoutine.maxAdds = Math.Max(1, WPvPRoutine.maxAdds-1);
                }
            }
            else
            {
                WPvPRoutine.Debug("Oracle needs {0} more combats", WINDOW - manaPercentsAfterCombats.Count);
                WPvPRoutine.maxAdds = 1;
            }

            WPvPRoutine.Debug("Oracle set maxAdds to " + WPvPRoutine.maxAdds);
        }

    }
}
