﻿#region Credits
/*
 * Warlock PvP (destruction at this moment)
 * Author: Dador
 * I take some paths from PrinceOfDarkness. Thanks to cowdude.
 */
#endregion
using System;
using System.Threading;
using System.Collections.Generic;
using System.Linq;
using Timer = System.Timers.Timer;

using Styx;
using Styx.Common;
using Styx.Combat.CombatRoutine;
using System.Windows.Media;
using Styx.CommonBot;
using Styx.CommonBot.Routines;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;

namespace WPvP
{
    public partial class WPvPRoutine : CombatRoutine
    {
        public static readonly Version Version = new Version(1, 0);

        public override string Name {get {return "Warlock PvP " + Version;} }

        public override WoWClass Class { get { return WoWClass.Warlock; } }

        public static LocalPlayer Me { get { return ObjectManager.Me; } }

        public static void Log(string msg, params object[] args)
        {
            if (msg != null)
            {
                Logging.Write(LogLevel.Normal, Colors.DarkOrange, "[WPvP]: " + msg, args);
            }
        }

        public static void Debug(string msg, params object[] args)
        {
            if (msg != null)
            {
                Logging.Write(LogLevel.Verbose, Colors.DarkOrange, "[WPvP Debug]: " + msg, args);
            }
        }

        public int CountOfBurningEmbers
        {
            get
            {
                return Lua.GetReturnVal<int>("return UnitPower('player', 14)", 0);
            }
        }

        public override void Initialize()
        {
            Log("Stating {0}", Name);
            Debug("Init: binding lua events");
            Lua.Events.AttachEvent("UNIT_SPELLCAST_SUCCEEDED", Spells.HandleSpellSucceeded);
            Lua.Events.AttachEvent("UNIT_SPELLCAST_STOP", Spells.HandleSpellFailure);
            Lua.Events.AttachEvent("UNIT_SPELLCAST_FAILED", Spells.HandleSpellFailure);
            Lua.Events.AttachEvent("UI_ERROR_MESSAGE", HandleErrorMessage);
            Lua.Events.AttachEvent("PLAYER_REGEN_ENABLED", HandleCombatDone);
            Lua.Events.AttachEvent("PET_BAR_UPDATE", FelPuppy.HandlePetBarUpdated);

            Debug("Getting lua vars");
            SPELL_FAILED_UNIT_NOT_INFRONT = Lua.GetReturnVal<string>("return SPELL_FAILED_UNIT_NOT_INFRONT", 0);

            //handle pet spawn delay on dismount
            MountHandler.OnDismount += delegate
            {
                Debug("Resetting checkpet timer since player dismounted - allow calls in at least 2 seconds.");
                PetCheckTimer = DateTime.Now.AddSeconds(2);
            };

            Debug("Binding bot events...");
            BotEvents.OnBotStarted += delegate(EventArgs args)
            {
                ObjectManager.Update();
                FelPuppy.UpdatePetSkills();
                Spells.ForceUnlock();
                Behavior.Reset();
                Log("{0} is ready for action! Your warlock spec is: {1}", Name, Specialization.ToString());
            };

            BotEvents.OnBotStopped += delegate(EventArgs args)
            {
                //clean anything needed
                FelPuppy.RemoveAllAutoCast();
                Spells.ForceUnlock();
            };
        }
        
        public override bool NeedRest
        {
            get
            {
                MountHandler.Pulse();

                if (Me.Mounted)
                    return false;
                if (Me.IsDead || Me.IsGhost || Me.Combat)
                    return false;

                if (Me.HealthPercent < 50 || Me.ManaPercent < 60)
                {
                    Debug("NeedRest is true");
                    return true;
                }
                
                return false;
            }
        }


        public static DateTime PetCheckTimer = DateTime.MinValue; //prevent pet actions right after dismounting
        public static int maxAdds = 1; //see behavior.cs to edit this

        public override bool NeedPreCombatBuffs
        {
            get
            {
                //return false; // TODO: check what happens
                MountHandler.Pulse();

                if (Me.Mounted)
                    return false;

                if (CheckPet(true))
                    return true;

                if (Me.CastingSpellId == Spells.CreateHealthstone.Id || Me.CastingSpellId == Spells.SoulStone.Id)
                {
                    Debug("Need buffs: creating hearthstone or soulstone");
                    return true;
                }

                if (Spells.PlayerNeedsBuff(Spells.SoulLink))
                {
                    Debug("Need buffs: no soul link");
                    return true;
                }

                if (NeedsHealthstone)
                {
                    Debug("Need buffs: no healthstone");
                    return true;
                }

                if (!UnitHelper.HasBuff(Me, Spells.SoulStone) && Spells.CanCast(Spells.SoulStone))
                {
                    Debug("Need buffs: soulstone is ready to use");
                    return true;
                }

                return false;
            }
        }

        private string SPELL_FAILED_UNIT_NOT_INFRONT = null;
        private void HandleErrorMessage(object sender, LuaEventArgs args)
        {
            string msg = args.Args[0].ToString();

            if (msg == SPELL_FAILED_UNIT_NOT_INFRONT)
                FaceTarget();
        }

        private void HandleCombatDone(object sender, LuaEventArgs args)
        {
            Debug("--- Combat done! ---");
            Behavior.AddCombatStats();
        }

        public static void FaceTarget()
        {
            if (Me.CurrentTarget != null)
            {
                Debug("Correcting target facing.");
                Me.CurrentTarget.Face();
            }
        }


        public override void Rest()
        {
            MountHandler.Pulse();

            //a bit ugly, but well... as long as it works.
            if (Spells.Locked 
                || Me.IsCasting
                || Me.ActiveAuras.Any(a => a.Value.Name == "Food" || a.Value.Name == "Drink"))
            {
                //Debug("Rest: do nothing, already busy");
                return;
            }

            if (Me.HealthPercent < 50 || Me.ManaPercent < 60)
            {
                Debug("Resting lazily with drink/food...");

                if (Me.HealthPercent < 50)
                    Styx.CommonBot.Rest.FeedImmediate();
                if (Me.ManaPercent < 60)
                    Styx.CommonBot.Rest.DrinkImmediate();

                //hb bug detecting auras
                Thread.Sleep(1000);
            }
        }



        public override void PreCombatBuff()
        {
            MountHandler.Pulse();

            // Wait for us to stop casting please...
            if (Spells.Locked || Me.IsCasting)
            {
                Debug("Precombat buffs: do nothing as I'm casting...");
                return;
            }

            //pet
            if (CheckPet(true))
            {
                Debug("Precombat: need a pet!");
                return;
            }

            if (Spells.PlayerNeedsBuff(Spells.SoulLink))
            {
                Debug("Precombat: need soul link");
                Spells.Cast(Spells.SoulLink);
                return;
            }

            if (NeedsHealthstone)
            {
                Debug("Precombat: need a healthstone");
                Spells.Cast(Spells.CreateHealthstone);
                return;
            }

            if (!UnitHelper.HasBuff(Me, Spells.SoulStone) && Spells.CanCast(Spells.SoulStone))
            {
                Debug("Precombat: get stoned (got it?)");
                Spells.Cast(Spells.SoulStone);
                return;
            }
        }



        public override void Pull()
        {
            MountHandler.Pulse();

            if (Spells.Locked || Me.IsCasting)
            {
                return;
            }

            //hb bug - it sometimes call Pull() right after killing a target without looting/resting
            if (!Me.Combat)
                FelPuppy.Attack();

            Spells.Cast(Opener, Me.CurrentTarget);
        }

        public override void Combat()
        {
            //pulses whatever needs it.
            FelPuppy.Pulse();
            MountHandler.Pulse();

            if (Me.CurrentTarget == null || Me.CurrentTarget.IsDead)
            {
                Debug("WARN: Me.Target is null/dead in Combat()");
                if (Targeting.Instance.FirstUnit == null || Targeting.Instance.FirstUnit.IsDead || !Targeting.Instance.FirstUnit.Combat)
                {
                    // Just try to pick up our pets target
                    if (Me.Pet != null && Me.Pet.Combat && Me.Pet.CurrentTarget != null)
                    {
                        Debug("Combat: picked up pet target");
                        TargetUnit(Me.Pet.CurrentTarget);
                        return;
                    }

                    //this seems to be ok... happens when combat done. still in combat but target died
                    //Debug("Combat: oops, no target!");
                    return;
                }
                TargetUnit(Targeting.Instance.FirstUnit);
                if (Me.CurrentTarget != null)
                    Debug("Picked up new target: " + Me.CurrentTarget.Name);
                return;
            }

            // Don't do anything while casting!!
            if (Me.IsCasting || Spells.Locked)
            {
                // Cast Fel Flame right after Incinerate
                if (Me.CastingSpell != null &&
                    Me.CastingSpell.Id == Spells.Incinerate.Id &&
                    Me.CurrentCastTimeLeft != null &&
                    Me.CurrentCastTimeLeft <= new TimeSpan(0, 0, 1))
                {
                    Debug("Cast Fel Flame right after Incinerate");
                    Spells.FelFlame.Cast();
                }
                //Debug("Combat: do nothing, casting...");
                if (!Me.IsFacing(Me.CurrentTarget))
                {
                    FaceTarget();
                }
                return;
            }

            if (Me.IsMoving)
            {
                Styx.WoWInternals.WoWMovement.MoveStop();
            }

            if (CheckPet(false))
            {
                Debug("Stopping Combat - CheckPet=true");
                return;
            }

            //if low life, try casting voidie's bubble
            if (Me.HealthPercent < 50 && FelPuppy.HasSpell(Spells.VoidWalker_Sacrifice)) // TODO: check in game
            {
                FelPuppy.Cast(Spells.VoidWalker_Sacrifice);
            }

            // Aggro management with more than one add
            var adds = ObjectManager.GetObjectsOfType<WoWUnit>(true, false).
                Where(u => u.CurrentTargetGuid == Me.Guid && u.Distance < 50 && u.IsHostile).OrderBy(u => u.Guid).ToList();
            if (adds.Count > 0)
            {
                Debug("I have {0} bad guys on me", adds.Count);

                //help me, stupid pet
                if (FelPuppy.IsValid
                    && Me.Pet.CurrentTarget != null 
                    && Me.Pet.CurrentTarget.CurrentTarget == Me.Pet)
                {
                    Debug("Sending pet on add");
                    var oldTarget = Me.CurrentTarget;
                    TargetUnit(adds[0]);

                    //death coil add
                    if (Spells.CanCast(Spells.DeathCoil))
                    {
                        Debug("Death coil on add");
                        Spells.Cast(Spells.DeathCoil, adds[0]);
                    }

                    FelPuppy.Attack();
                    TargetUnit(oldTarget);
                    return;
                }

                //too many bad guys?
                if (adds.Count > maxAdds)
                {
                    Debug("WARN: Too many adds on me!");

                    //compute average adds pos. etc.
                    WoWPoint middle = WoWPoint.Zero;
                    int nearFoes = 0;
                    foreach (var add in adds)
                    {
                        middle.Add(add.Location.X / adds.Count,
                                    add.Location.Y / adds.Count,
                                    add.Location.Z / adds.Count);
                        if (add.DistanceSqr < 100)
                            nearFoes++;
                    }

                    //try soulshatter if we got a pet
                    if (!PVPLogicEnabled
                        && Me.HealthPercent < 65 
                        && FelPuppy.IsValid
                        && Spells.CanCast(Spells.Soulshatter))
                    {
                        Spells.Cast(Spells.Soulshatter);
                        return;
                    }

                    //pvp/im-gonna-die aoe fear!
                    if ((PVPLogicEnabled || Me.CurrentHealth < 30) 
                        && nearFoes > 1
                        && Spells.CanCast(Spells.HowlOfTerror))
                    {
                        Debug("Popping howl of terror");
                        Spells.Cast(Spells.HowlOfTerror);
                        return;
                    }

                    /*
                    //or pop an infernal. yay.
                    if (Spells.SummonInfernal.CanCast
                        && (!PVPLogicEnabled || (PVPLogicEnabled && Me.CurrentHealth > 70))
                        && middle.Distance(Me.Location) < Spells.SummonInfernal.MaxRange-8)
                    {
                        Spells.Cast(Spells.SummonInfernal, middle); //fix this
                        return;
                    }*/
                }
            }

            if (Me.HealthPercent <= 40 && PlayerHasItem(HEALTHSTONE) && Spells.CanCast(Spells.UseHealthStone))
            {
                Debug("Use healthstone");
                Spells.Cast(Spells.UseHealthStone);
                return;
            }

            if (Me.HealthPercent >= 40 && Me.Pet != null && Me.Pet.HealthPercent < 60 && Spells.CanCast(Spells.HealthFunnel))
            {
                Spells.Cast(Spells.HealthFunnel);
                return;
            }

            /* 
             * destruction combat rotation
             * ***************************
             * - pull (Immolate)
             * - Conflagrate
             * - Fel Flame and Incinerate right after Fel Flame
             * - when have Burning Embers, then cast Chaos Bolt or Shadowburn (when target HP low 20%)
             */

            //pet attack
            if (FelPuppy.IsValid && Me.Pet.CurrentTarget == null)
            {
                Debug("Send pet on target.");
                FelPuppy.Attack();
            }

            //finish spell
            if (Me.CurrentTarget.HealthPercent < 20 && CountOfBurningEmbers>0 && Spells.CanCast(Spells.Shadowburn))
            {
               Debug("Low HP target, so cast Shadowburn!");
               Spells.Cast(Spells.Shadowburn, Me.CurrentTarget);
               return;
            }

            // Immolate
            if (Spells.CanCast(Spells.CurseOfTheElements)
                && UnitHelper.MustRefreshBuff(Me.CurrentTarget, Spells.CurseOfTheElements))
            {
                Debug("Target needs Curse of The Elements");
                Spells.Cast(Spells.CurseOfTheElements, Me.CurrentTarget);
            }

            if (Spells.CanCast(Spells.Conflagrate))
            {
                Debug("Cast Conflagrate");
                Spells.Cast(Spells.Conflagrate, Me.CurrentTarget);
                Thread.Sleep(SpellManager.GlobalCooldownLeft);
            }

            // Immolate
            if (Spells.CanCast(Spells.Immolate)
                && UnitHelper.MustRefreshBuff(Me.CurrentTarget, Spells.Immolate))
            {
                Debug("Target needs Immolate");
                Spells.Cast(Spells.Immolate, Me.CurrentTarget);
                return;
            }

            if (CountOfBurningEmbers > 0 && Spells.CanCast(Spells.ChaosBolt))
            {
                Debug("We have burning embers, so cast Chaos Bolt");
                Spells.Cast(Spells.ChaosBolt, Me.CurrentTarget);
                return;
            }
            /*else if (Spells.CanCast(Spells.FelFlame))
            {
                Debug("We need burning embers, so cast Fel Flame");
                Spells.Cast(Spells.FelFlame, Me.CurrentTarget);
                return;
            }*/
            else if (Spells.CanCast(Spells.Incinerate))
            {
                Debug("We need burning embers, but can't cast Fel Flame, so cast Incinerate");
                Spells.Cast(Spells.Incinerate, Me.CurrentTarget);
                return;
            }

            //Debug("Try Cast Conflagrate");
            //Spells.Conflagrate.Cast();

            //pvp fear
            if (PVPLogicEnabled && Spells.CanCast(Spells.Fear))
            {
                List<WoWPlayer> players = ObjectManager.GetObjectsOfType<WoWPlayer>(false, false).
                    Where(p =>
                                p.DistanceSqr < Math.Pow(Spells.Fear.MaxRange, 2) - 10
                                && !p.Mounted
                                && !UnitHelper.HasBuff(p, Spells.Fear)).
                    OrderBy(p => p.DistanceSqr).
                    ToList();
                WoWPlayer fearme = null;
                if (players.Count > 0)
                    fearme = players.Count > 1 ? players[1] : players[0]; // TODO: check on current target

                if (fearme != null)
                {
                    Debug("PVP Logic: fearing player {0}. Distance={1}, FearRange={2}",
                        fearme.Name, fearme.Distance, Spells.Fear.MaxRange);
                    Spells.Cast(Spells.Fear, fearme);
                    return;
                }
            }

            // Moving somewhere

            /*
            //nothing to do ?! desperate attack.
            if (Me.CurrentMana < 5 && Spells.CanCast(Spells.Shoot))
            {
                Debug("Dont know what to cast... shooting");
                Spells.Cast(Spells.Shoot, Me.CurrentTarget);
            }*/

            //Debug("Did nothing in Combat() loop. Slow FPS? client lag?");
        }

        public void RandomMoving()
        {
            if (Me.CurrentTarget != null)
            {
                if (Me.CurrentTarget.MovementInfo.IsMoving)
                {

                }
                else
                {
                    if (Me.CurrentTarget.Distance > 20)
                    {
                        Debug("My Position: {0}", Me.Location);
                        Debug("Target Position: {0}", Me.CurrentTarget.Location);
                        Debug("Calculated Location: ", Me.CurrentTarget.Location.GetDirectionTo(Me.Location));
                    }
                }
            }
        }


        //should we use special pvp spells like fear?
        public static bool PVPLogicEnabled
        {
            get
            {
                return
                    Battlegrounds.IsInsideBattleground
                    || (Me.CurrentTarget != null 
                        && Me.CurrentTarget.IsPlayer 
                        && Me.CurrentTarget.IsTargetingMeOrPet 
                        && Me.CurrentTarget.IsHostile);
            }
        }

        /**************************************************************************/

        /* Lock helpers */

        public static uint HEALTHSTONE { get { return 5512; } }

        public static WoWSpell Opener
        {
            get
            {
                var wish = new WoWSpell[] {
                    Spells.Immolate,
                    Spells.FelFlame,
                    Spells.Incinerate
                };

                WoWSpell spell = wish.First(s => s.CanCast);
                Debug("Opener: " + spell.Name);
                return spell;
            }
        }

        public static int TotalAdds
        {
            get
            {
                return ObjectManager.GetObjectsOfType<WoWUnit>(true, false).
                    Where(u => u.Attackable && (u.CurrentTargetGuid == Me.Guid || (Me.Pet != null && u.CurrentTargetGuid == Me.Pet.Guid))).
                    Count();
            }
        }

        public static bool NeedsHealthstone
        {
            get
            {
                return
                    Me.BagItems != null && Me.BagItems.Count > 0 && //bug
                    !PlayerHasItem(HEALTHSTONE)
                    && SpellManager.HasSpell(Spells.CreateHealthstone);
            }
        }

        public static bool PlayerHasItem(WoWItem item)
        {
            return PlayerHasItem(item.Entry);
        }
        public static bool PlayerHasItem(uint entry)
        {
            /*return
                ObjectManager.GetObjectsOfType<WoWItem>(false, false).
                    Exists(i => i.Entry == entry);*/
            return Me.BagItems.Exists(i => i.Entry == entry);
        }

        public static WoWItem FindItem(uint entry)
        {
            /*return
                ObjectManager.GetObjectsOfType<WoWItem>(false, false).
                    FirstOrDefault(i => i.Entry == entry);*/
            return Me.BagItems.FirstOrDefault(i => i.Entry == entry);
        }

        public static void TargetUnit(WoWUnit unit)
        {
            if (unit == null || unit.Guid == Me.CurrentTargetGuid)
                return;
            Debug("Target unit " + unit.Name);
            unit.Target();
            LuaEventWait evt = new LuaEventWait("PLAYER_TARGET_CHANGED");
            evt.Wait(500);
            Debug("Successfuly changed player's target");
        }

        public enum Specializations
        {
            Affliction,
            Demonology,
            Destruction,
            None
        }

        public static Specializations Specialization
        {
            get
            {
                if (SpellManager.HasSpell(Spells.SPEC_Destruction))
                    return Specializations.Destruction;
                else
                    return Specializations.None;
            }
        }

        private static bool noPet = false; //skip pet actions or check?
        public static bool CheckPet(bool forceSummon)
        {
            //returns true if summoning a new pet

            //we're using a timer to prevent checking pet if we just dismounted
            //as it takes a few seconds to see the pet spawning...
            if (DateTime.Now.CompareTo(PetCheckTimer) >= 0)
            {
                Debug("CheckPet timer has ended. Allow calling CheckPet now");
                //reset checkpet timer
                Debug("Resetting checkpet timer - allow calls in at least 10 seconds.");
                PetCheckTimer = DateTime.Now.AddSeconds(10);
            }
            else
            {
                //Debug("Preventing checking pet for a few seconds...");
                return false;
            }

            if (!noPet && !FelPuppy.IsValid)
            {
                Debug("I need a new puppy.");
                
                noPet = false;

                if (Spells.CanCast(Spells.FlamesOfXoroth))
                {
                    Spells.Cast(Spells.FlamesOfXoroth);
                }
                else
                {
                    if (SpellManager.HasSpell(Spells.SummonFelhunter))
                        Spells.Cast(Spells.SummonFelhunter);
                    else if (SpellManager.HasSpell(Spells.SummonVoidwalker))
                        Spells.Cast(Spells.SummonVoidwalker);
                    else if (SpellManager.HasSpell(Spells.SummonImp))
                        Spells.Cast(Spells.SummonImp);
                    else
                    {
                        Debug("Hey... no pet found?!"); // this should never happen
                        noPet = true;
                        return false;
                    }
                }
                return true;
            }
            return false;
        }
    }
}